"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const option_1 = require("@reactive-js/option");
const pipe_1 = require("@reactive-js/pipe");
class CharStreamImpl {
    constructor(src) {
        this.src = src;
        this._index = -1;
        this.current = -1;
        this.hasCurrent = false;
    }
    get index() {
        return this._index;
    }
    set index(index) {
        this._index = index;
    }
    move(_) {
        this.hasCurrent = false;
        this.current = -1;
        this._index++;
        const index = this._index;
        const src = this.src;
        if (this._index < src.length) {
            this.hasCurrent = true;
            this.current = src.charCodeAt(index);
        }
        return this.hasCurrent;
    }
}
class ParserError {
    constructor(index) {
        this.index = index;
        this.error = new Error();
    }
    get stack() {
        return this.error.stack;
    }
}
const throwParseErrorDev = (charStream) => {
    const error = new ParserError(charStream.index);
    throw error;
};
const parseErrorSymbol = Symbol("ParseError");
const throwParseErrorProd = (_) => {
    throw parseErrorSymbol;
};
exports.throwParseError = process.env.NODE_ENV === "production"
    ? throwParseErrorProd
    : throwParseErrorDev;
const isParseErrorDev = (e) => e instanceof ParserError;
const isParseErrorProd = (e) => e === parseErrorSymbol;
exports.isParseError = process.env.NODE_ENV === "production" ? isParseErrorProd : isParseErrorDev;
exports.createCharStream = (input) => new CharStreamImpl(input);
function concat(...parsers) {
    return charStream => {
        const result = [];
        for (const parse of parsers) {
            const next = parse(charStream);
            result.push(next);
        }
        return result;
    };
}
exports.concat = concat;
exports.followedBy = (other) => parser => charStream => {
    const result = parser(charStream);
    other(charStream);
    return result;
};
exports.map = (mapper) => parser => pipe_1.compose(parser, mapper);
exports.mapTo = (v) => {
    const mapper = (_) => v;
    return exports.map(mapper);
};
exports.parseWithOrThrow = (parser) => {
    const parse = pipe_1.pipe(parser, exports.followedBy(exports.pEof));
    return input => {
        const charStream = exports.createCharStream(input);
        return parse(charStream);
    };
};
exports.parseWith = (parse) => {
    const doParse = exports.parseWithOrThrow(parse);
    return input => {
        try {
            return doParse(input);
        }
        catch (e) {
            if (exports.isParseError(e)) {
                return option_1.none;
            }
            throw e;
        }
    };
};
exports.or = (otherParse) => parse => charStream => {
    const index = charStream.index;
    try {
        return parse(charStream);
    }
    catch (e) {
        if (exports.isParseError(e)) {
            charStream.index = index;
            return otherParse(charStream);
        }
        else {
            throw e;
        }
    }
};
exports.many = (options = {}) => parse => charStream => {
    const { min = 0, max = Number.MAX_SAFE_INTEGER } = options;
    const retval = [];
    let count = 0;
    let index = -1;
    try {
        while (count < max) {
            index = charStream.index;
            const next = parse(charStream);
            count++;
            retval.push(next);
        }
    }
    catch (e) {
        if (exports.isParseError(e)) {
            charStream.index = index;
        }
        else {
            throw e;
        }
    }
    return count < min ? exports.throwParseError(charStream) : retval;
};
exports.manyIgnore = (options = {}) => parse => charStream => {
    const { min = 0, max = Number.MAX_SAFE_INTEGER } = options;
    let count = 0;
    let index = -1;
    try {
        while (count < max) {
            index = charStream.index;
            parse(charStream);
            count++;
        }
    }
    catch (e) {
        if (exports.isParseError(e)) {
            charStream.index = index;
        }
        else {
            throw e;
        }
    }
    return count < min ? exports.throwParseError(charStream) : option_1.none;
};
exports.optional = (parse) => charStream => {
    const index = charStream.index;
    try {
        return parse(charStream);
    }
    catch (e) {
        if (exports.isParseError(e)) {
            charStream.index = index;
            return option_1.none;
        }
        else {
            throw e;
        }
    }
};
exports.orDefault = (default_) => pipe_1.compose(exports.optional, exports.map(result => result !== null && result !== void 0 ? result : default_()));
exports.sepBy1 = (separator) => parser => {
    const parseTailValue = (charStream) => {
        separator(charStream);
        return parser(charStream);
    };
    const parseTail = exports.many()(parseTailValue);
    return charStream => {
        const first = parser(charStream);
        const tail = parseTail(charStream);
        tail.unshift(first);
        return tail;
    };
};
exports.sepBy = (separator) => pipe_1.compose(exports.sepBy1(separator), exports.orDefault(() => []));
exports.string = (str) => charStream => {
    charStream.move();
    const match = charStream.src.startsWith(str, charStream.index);
    if (match) {
        charStream.index += str.length - 1;
        return str;
    }
    else {
        return exports.throwParseError(charStream);
    }
};
exports.satisfy = (f) => charStream => {
    if (charStream.move()) {
        const current = charStream.current;
        if (f(current)) {
            return current;
        }
    }
    return exports.throwParseError(charStream);
};
exports.manySatisfy = (options = {}) => parser => {
    const parse = exports.manyIgnore(options)(parser);
    return charStream => {
        const start = charStream.index + 1;
        parse(charStream);
        return charStream.src.substring(start, charStream.index + 1);
    };
};
exports.char = (c) => {
    const charCode = c.charCodeAt(0);
    return exports.satisfy(x => x === charCode);
};
exports.pEof = (charStream) => charStream.move() ? exports.throwParseError(charStream) : option_1.none;
exports.pSemicolon = exports.char(";");
exports.pComma = exports.char(",");
exports.pSpace = exports.char(" ");
exports.pColon = exports.char(":");
exports.pPeriod = exports.char(".");
exports.pEquals = exports.char("=");
exports.pForwardSlash = exports.char("/");
exports.pDash = exports.char("-");
exports.pOpenParen = exports.char("(");
exports.pCloseParen = exports.char(")");
exports.pDquote = exports.char('"');
exports.pAsterisk = exports.char("*");
//# sourceMappingURL=index.js.map