import { none } from "@reactive-js/option";
import { compose, pipe } from "@reactive-js/pipe";
class CharStreamImpl {
    constructor(src) {
        this.src = src;
        this._index = -1;
        this.current = -1;
        this.hasCurrent = false;
    }
    get index() {
        return this._index;
    }
    set index(index) {
        this._index = index;
    }
    move(_) {
        this.hasCurrent = false;
        this.current = -1;
        this._index++;
        const index = this._index;
        const src = this.src;
        if (this._index < src.length) {
            this.hasCurrent = true;
            this.current = src.charCodeAt(index);
        }
        return this.hasCurrent;
    }
}
class ParserError {
    constructor(index) {
        this.index = index;
        this.error = new Error();
    }
    get stack() {
        return this.error.stack;
    }
}
const throwParseErrorDev = (charStream) => {
    const error = new ParserError(charStream.index);
    throw error;
};
const parseErrorSymbol = Symbol("ParseError");
const throwParseErrorProd = (_) => {
    throw parseErrorSymbol;
};
export const throwParseError = process.env.NODE_ENV === "production"
    ? throwParseErrorProd
    : throwParseErrorDev;
const isParseErrorDev = (e) => e instanceof ParserError;
const isParseErrorProd = (e) => e === parseErrorSymbol;
export const isParseError = process.env.NODE_ENV === "production" ? isParseErrorProd : isParseErrorDev;
export const createCharStream = (input) => new CharStreamImpl(input);
export function concat(...parsers) {
    return charStream => {
        const result = [];
        for (const parse of parsers) {
            const next = parse(charStream);
            result.push(next);
        }
        return result;
    };
}
export const followedBy = (other) => parser => charStream => {
    const result = parser(charStream);
    other(charStream);
    return result;
};
export const map = (mapper) => parser => compose(parser, mapper);
export const mapTo = (v) => {
    const mapper = (_) => v;
    return map(mapper);
};
export const parseWithOrThrow = (parser) => {
    const parse = pipe(parser, followedBy(pEof));
    return input => {
        const charStream = createCharStream(input);
        return parse(charStream);
    };
};
export const parseWith = (parse) => {
    const doParse = parseWithOrThrow(parse);
    return input => {
        try {
            return doParse(input);
        }
        catch (e) {
            if (isParseError(e)) {
                return none;
            }
            throw e;
        }
    };
};
export const or = (otherParse) => parse => charStream => {
    const index = charStream.index;
    try {
        return parse(charStream);
    }
    catch (e) {
        if (isParseError(e)) {
            charStream.index = index;
            return otherParse(charStream);
        }
        else {
            throw e;
        }
    }
};
export const many = (options = {}) => parse => charStream => {
    const { min = 0, max = Number.MAX_SAFE_INTEGER } = options;
    const retval = [];
    let count = 0;
    let index = -1;
    try {
        while (count < max) {
            index = charStream.index;
            const next = parse(charStream);
            count++;
            retval.push(next);
        }
    }
    catch (e) {
        if (isParseError(e)) {
            charStream.index = index;
        }
        else {
            throw e;
        }
    }
    return count < min ? throwParseError(charStream) : retval;
};
export const manyIgnore = (options = {}) => parse => charStream => {
    const { min = 0, max = Number.MAX_SAFE_INTEGER } = options;
    let count = 0;
    let index = -1;
    try {
        while (count < max) {
            index = charStream.index;
            parse(charStream);
            count++;
        }
    }
    catch (e) {
        if (isParseError(e)) {
            charStream.index = index;
        }
        else {
            throw e;
        }
    }
    return count < min ? throwParseError(charStream) : none;
};
export const optional = (parse) => charStream => {
    const index = charStream.index;
    try {
        return parse(charStream);
    }
    catch (e) {
        if (isParseError(e)) {
            charStream.index = index;
            return none;
        }
        else {
            throw e;
        }
    }
};
export const orDefault = (default_) => compose(optional, map(result => result !== null && result !== void 0 ? result : default_()));
export const sepBy1 = (separator) => parser => {
    const parseTailValue = (charStream) => {
        separator(charStream);
        return parser(charStream);
    };
    const parseTail = many()(parseTailValue);
    return charStream => {
        const first = parser(charStream);
        const tail = parseTail(charStream);
        tail.unshift(first);
        return tail;
    };
};
export const sepBy = (separator) => compose(sepBy1(separator), orDefault(() => []));
export const string = (str) => charStream => {
    charStream.move();
    const match = charStream.src.startsWith(str, charStream.index);
    if (match) {
        charStream.index += str.length - 1;
        return str;
    }
    else {
        return throwParseError(charStream);
    }
};
export const satisfy = (f) => charStream => {
    if (charStream.move()) {
        const current = charStream.current;
        if (f(current)) {
            return current;
        }
    }
    return throwParseError(charStream);
};
export const manySatisfy = (options = {}) => parser => {
    const parse = manyIgnore(options)(parser);
    return charStream => {
        const start = charStream.index + 1;
        parse(charStream);
        return charStream.src.substring(start, charStream.index + 1);
    };
};
export const char = (c) => {
    const charCode = c.charCodeAt(0);
    return satisfy(x => x === charCode);
};
export const pEof = (charStream) => charStream.move() ? throwParseError(charStream) : none;
export const pSemicolon = char(";");
export const pComma = char(",");
export const pSpace = char(" ");
export const pColon = char(":");
export const pPeriod = char(".");
export const pEquals = char("=");
export const pForwardSlash = char("/");
export const pDash = char("-");
export const pOpenParen = char("(");
export const pCloseParen = char(")");
export const pDquote = char('"');
export const pAsterisk = char("*");
//# sourceMappingURL=index.js.map